#!/usr/bin/perl

use strict;
use IO::Zlib;

## read in ld
my $ldFile = "DEBUG/test.h5.ld";
my $R0 = 0.1;
my $vcfFile = "DEBUG/chr20.filtered.vcf.MAF0.05.h5.vcf";
my $subjCol = 9;
my $dir = ".";
my $vcfFile0 = "";
my $key = "";
my $depthKey = "";

if (scalar(@ARGV) < 7){
    die "Usage [File for paired Rsquare][Rsquare Cutoff (0.1)][Vcf File][Subj Column][Output file prefix][key for genotype likelihood (GL/PL)] [key for coverage (DP/GD)]\n";
}else{
    $ldFile = $ARGV[0];
    $R0 = $ARGV[1];
    $vcfFile0 = $ARGV[2];
    $subjCol =$ARGV[3];
    $dir = $ARGV[4];
    $key = $ARGV[5];
    $depthKey = $ARGV[6];
}

$vcfFile = "$vcfFile0";
my $outFile = "$dir.out.sub$subjCol.R$R0.txt";
my $outFile2 = "$dir.out.sub$subjCol.R$R0.pruned.vcf";
print "writing to file $outFile2\n";

## read in the vcf file
if ($vcfFile =~ /\.gz$/){
    tie (*VCF, "IO::Zlib",$vcfFile,"rb") || die "Cannot open file $vcfFile to read\n";
    print "Openning bgziped vcf file $vcfFile ...\n";
} else{
    open(VCF, $vcfFile) || die "Cannot open $vcfFile to read\n";
    print "Openning vcf file $vcfFile ...\n";
}
## take a peak at the vcfFile and figure out where is the GL or PL key
my $lineCount = 0;
my $idxGLK = 999;
my $idxDP = 999;
while(<VCF>){
    if (/^#/){
        next;
    }else{
        $lineCount ++;
        my @F = split();
        my $format = $F[8];
        my @formatKeys = split(/:/,$F[8]);
        my $lastIdx = $#formatKeys;
        $idxGLK = $lastIdx;
        $idxDP = $lastIdx;
        for (my $i = 0; $i < scalar(@formatKeys); $i ++){
            if ($formatKeys[$i] eq $key){
                $idxGLK = $i;
            }
            if ($formatKeys[$i] eq $depthKey){
                $idxDP = $i;
            }
            #print "GLK key $key, idx $idxGLK, DP key $depthKey, idx $idxDP ...\n";
        }

        if ($idxGLK == $lastIdx and $formatKeys[$lastIdx] ne $key){
          die "Incorrect key ($key) for genotype likelihood field in ".$F[8]."\n";
        }
        if ($idxDP == $lastIdx and $formatKeys[$lastIdx] ne $depthKey){
          die "Incorrect key ($depthKey) for depth field in ".$F[8]."\n";
        }

        print "the genotype likelihood in ".$F[8]." is $idxGLK \($formatKeys[$idxGLK]\) \n";
        print "the coverage in ".$F[8]." is $idxDP \($formatKeys[$idxDP]\) \n";
        if ($lineCount >= 1){
            last;
        }
    }
}
close VCF;

## creat a hash to record the information of the sparse matrix
my %Rpairs = ();
open (LD,$ldFile) || die "Cannot read in $ldFile to read\n";
print "Reading in the LD information from the file $ldFile ... \n";

while (<LD>){
    if ($. == 1 && /ChrA/){next;}
    my @F = split();
    my $v0 = $F[0].":".$F[1];
    my $v1 = $F[2].":".$F[3];
    my $R = $F[4];
    if (abs($R) > $R0){
        ## put v0 in the hash
        if (!exists($Rpairs{$v0})){
          $Rpairs{$v0} = $v1;
        }else{
            $Rpairs{$v0} .= ",".$v1; 
        }
        ## put v1 in the hash
        if (!exists($Rpairs{$v1})){
          $Rpairs{$v1} = $v0;
        }else{
            $Rpairs{$v1} .= ",".$v0; 
        }
    }
}
close LD;

print "There are ".keys(%Rpairs)." keys in the Rpairs hash"."\n";


if ($vcfFile =~ /\.gz$/){
    tie (*VCF, "IO::Zlib",$vcfFile,"rb") || die "Cannot open file $vcfFile to read\n";
    print "Openning bgziped vcf file $vcfFile ...\n";
} else{
    open(VCF, $vcfFile) || die "Cannot open $vcfFile to read\n";
    print "Openning vcf file $vcfFile ...\n";
}

my %cvgHash = ();
my %glfHash = ();
my @allsites = ();

while (<VCF>){
    if (/^#/) {next;}
    my @F = split();
    my $v0 = $F[0].":".$F[1];
    push(@allsites,$v0);
    my $geno = $F[$subjCol];
    my @genos = split(/:/,$geno);
    $cvgHash{$v0} = $genos[$idxDP]; ## this need to be changed according to the vcf format
    $glfHash{$v0} = $genos[$idxGLK];
    my @testifGLF = split(/,/,$genos[$idxGLK]);
    #print join("\t",@F);
    #print join("\t",@testifGLF);
    if (scalar(@testifGLF) < 3){
        die "Wrong GLF column in the genotype field\n";
    }
}
close VCF;
print "There are ".keys(%cvgHash)." in the cvg hash"."\n";

## sort the variant by their coverage and carry on exclusion of keys
## then build the out list
my %outList = ();
foreach my $key (sort {$cvgHash{$b} cmp $cvgHash{$a}} keys %cvgHash){
    if (!exists($outList{$key})){
        my @outs = split(/,/,$Rpairs{$key});
        foreach my $v0 (@outs){
        $outList{$v0} = "";
        }
    }
}
print "There are ".keys(%outList)." in the out hash\n";
#foreach my $key (keys (%outList)){
#    print $key."\n";
#}

## print out the variant, cvg, glfs
open(OUT,">$outFile") || die "cannot open $outFile to write\n";
open(OUT2,">$outFile2") || die "cannot open $outFile2 to write\n";

print OUT "chrPos\texclude\tcvg\tglf\tgeno$subjCol\n";
foreach my $key (@allsites){
    if (exists($outList{$key})){
      print OUT $key."\t1\t".$cvgHash{$key}."\t0,0,0\t.\/.:$cvgHash{$key}:0,0,0\n";
    } else{
      print OUT $key."\t0\t".$cvgHash{$key}."\t$glfHash{$key}\t.\/.:$cvgHash{$key}:$glfHash{$key}\n";;
    }
}
close OUT;

## print out the modified vcf file
if ($vcfFile =~ /\.gz$/){
    tie (*VCF, "IO::Zlib",$vcfFile,"rb") || die "Cannot open file $vcfFile to read\n";
    print "Openning bgziped vcf file...\n";
} else{
    open(VCF, $vcfFile) || die "Cannot open $vcfFile to read\n";
}

my $format = "GT:$depthKey:$key";

while(<VCF>){
    if (/^#/){
      my @F = split();
      print OUT2 join("\t",@F[0..8])."\t".$F[$subjCol]."\n";
      #print OUT2  $_;
      ## here need to modify to remove the individuals that are not included
    } else{
        my @F = split();
        my $chrPos = $F[0].":".$F[1];
        if (exists ($outList{$chrPos})){
            print OUT2 join("\t",@F[0..7])."\t".$format."\t.\/.:$cvgHash{$chrPos}:0,0,0\n";
        }else{
            print OUT2 join("\t",@F[0..7])."\t".$format."\t.\/.:$cvgHash{$chrPos}:$glfHash{$chrPos}\n";
        }
    }
   
}
close OUT2;
