/*
 *  GLFfromVCF.cpp
 *  LocalAncestry
 *
 *  Created by Youna Hu on 4/1/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "GLFfromVCF.h"
#include <iostream>

GLFfromVCF::GLFfromVCF(String &inFile, String gllkKey, int subjColumn, int Time){
	colChr = 0;
	colPos = 1;
	colRef = 3;
	colAlt = 4;
	colInd = subjColumn;
	colFormat = 8;
	GLLKkey = gllkKey;
	splitGL = -1; // Which column to split as genotype likelihood in the genotype entry 
	numStatesOneChain = 2;
	numChains = 2;
	numStates = numStatesOneChain*numChains;
	T = Time; // T = 4 as default
	
	// automate to see which one is GL,
	FILE *HapMapF = fopen(inFile.c_str(), "r");
	if (HapMapF == NULL){
		error("Error in opening %s!\n",inFile.c_str());
		return ;
	}
	String fileEntry;
	StringArray fileTokens;
	String formatEntry;
	StringArray formatTokens;
	int fileStep = 0;
	
	while (!feof(HapMapF)) {
		fileEntry.ReadLine(HapMapF);
		fileStep ++;
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",inFile.c_str(),fileStep);
		}
		if (fileEntry.Left(1) != "#") { //I assume this will be the line of the first variant
			// split the whole line into link tokens
			fileTokens.AddTokens(fileEntry, "\t");
			// take out the FORMAT token
			formatEntry = fileTokens[colFormat];
			// split the format entry into an array
			formatTokens.AddTokens(formatEntry, ":");
			printf("The FORMAT entry is %s (size %d).\n",formatEntry.c_str(),formatTokens.Length());
			for (int i = 0; i < formatEntry.Length(); i++){
				if(formatTokens[i] == GLLKkey){
					splitGL = i;
					break;
				}						
			}
			if (splitGL == -1) {
				error("The key %s is not specified in the FORMAT entry!\n",GLLKkey.c_str());
			}
			// once we have identified which column is the genotype likelihood, we can break out of file reading loop
			break;
		}
		fileEntry.Clear();
		fileTokens.Clear();
		formatEntry.Clear();
		formatTokens.Clear();
	}
	fclose(HapMapF);
}

void GLFfromVCF::UpdateGLFs(String & inFile){
	FILE *HapMapF = fopen(inFile.c_str(), "r");
	if (HapMapF == NULL){
		error("Error in opening %s!\n",inFile.c_str());
		return ;
	}
	int fileStep = 0;
	String fileEntry;
	StringArray fileTokens;
	String snpID;
	String genoItem;
	String genoGLwComma;
	StringArray genoGLs;
	std::vector<double> tmpGLs;
	// read in the file and update the SNP frequencies and minor alleles
	while (!feof(HapMapF)) {
		fileEntry.ReadLine(HapMapF);
		fileStep ++;
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",inFile.c_str(),fileStep);
		}
		// printf("first two charecters: %s\n",fileEntry.Left(2).c_str());
		if (fileEntry.Left(2) != "##") { // not process the meta information lines
			fileTokens.AddTokens(fileEntry, "\t");
			if (fileEntry.Left(2) == "#C") { // the line starts with #CHROM has the ID
				if (colInd > fileTokens.Length()) {
					error("The column for this subject is not included in this vcf!\n");
				} else{
					idSubj2Decomp = fileTokens[colInd];
					printf("Subject to decompose ( ID: %s ).\n",idSubj2Decomp.c_str());
				}
			} else {
				// variant IDs
				// variantID.push_back(fileTokens[colChr]+ ":" + fileTokens[colPos]);
				variantID.push_back(fileTokens[colPos].AsInteger());
				// variant reference allele
				refA.push_back(fileTokens[colRef][0]);
				// variant alternative allele
				altA.push_back(fileTokens[colAlt][0]);
				// get the genotype item entry
				genoItem = fileTokens[colInd];
				// break the entry by ":" and take out the genotype likelihood term
				genoGLs.AddTokens(genoItem,":");
				// break the gllks by ","
				genoGLwComma = genoGLs[splitGL];
				genoGLs.Clear();
				genoGLs.AddTokens(genoGLwComma,",");
				// std::cout << genoGLwComma.c_str() << std::endl;
				
				// the GL is in phred-scale genotype likelihoods Q = -10log_10(P)
				// so the p of regular scale is p = 10^{-Q/10}
				for (int i = 0; i < genoGLs.Length(); i ++) {
					tmpGLs.push_back( pow(10, -(double) genoGLs[i].AsInteger()/ (double)10) );
					//tmpGLs.push_back( (double) genoGLs[i].AsDouble() );
				} 
				gllk.push_back(tmpGLs);
				tmpGLs.clear();
			}
		}
		fileEntry.Clear();
		fileTokens.Clear();
		genoItem.Clear();
		genoGLs.Clear();
	}
	fclose(HapMapF);
	std::cout << "The size of variantID: " << variantID.size() << std::endl;
	std::cout << "The size of refA: " << refA.size() << std::endl;
	std::cout << "The size of altA: " << altA.size() << std::endl;
	
	/*for (int i = 0; i < 10; i ++) {
		for (int j = 0; j < gllk[i].size(); j++) {
			std::cout << gllk[i][j] <<  "\t";
		}
		std::cout << std::endl;
	}	*/
}

void GLFfromVCF::UpdateGenoFreqs(String & inFile1, String & inFile2, String & inGDistFile){
	RefPplns ppln1;
	RefPplns ppln2;
	ppln1.UpdateMAFs(inFile1);
	ppln2.UpdateMAFs(inFile2);
	ppln1.UpdateGDist(inGDistFile);
		
	std::cout << "update geno - size of ppln1 " << ppln1.allSNPfrq.size() << std::endl;
	std::cout << "update geno - size of ppln2 " << ppln2.allSNPfrq.size() << std::endl;
	std::cout << "update geno --size to decomposed " << gllk.size() << std::endl;
	std::cout << "update gDsit -- size of gDist " << ppln1.geneticD.size() << std::endl;
	
	int nSnpOverlapped = 0;
	
	/*
	for (std::map<String,double>::iterator i = ppln1.allSNPfrq.begin(); i != ppln1.allSNPfrq.end(); ++i) {
		std::cout << (*i).first.c_str() << "\t" << (*i).second << std::endl;
	}
	return;
	 */
	double ppln1MAF, ppln2MAF;
	
	// f1 is the frequency of alternative in ppln1, 
	// f2 is the frequency of the alternative in ppln2
	double f1, f2; 
	std::vector< double > f12s;
	for (int i = 0; i < variantID.size(); i ++) {
		//std::cout << (int) (ppln1.allSNPfrq.find(variantID[i]) != ppln1.allSNPfrq.end()) <<  (int) (ppln2.allSNPfrq.find(variantID[i]) != ppln2.allSNPfrq.end()) << std::endl;
		if ( ppln1.allSNPfrq.find(variantID[i]) != ppln1.allSNPfrq.end() && ppln2.allSNPfrq.find(variantID[i]) != ppln2.allSNPfrq.end() 
				&& ppln1.geneticD.find(variantID[i]) != ppln1.geneticD.end() ) {
			
			// put down the overlapped variant IDs
			overlapVariantID.push_back(variantID[i]);
			
			// store the MAF for ppln1 and ppln2
			ppln1MAF = ppln1.allSNPfrq[variantID[i]];
			ppln2MAF = ppln2.allSNPfrq[variantID[i]];
			
			// write down the allele frequency for the alternative alleles for the two reference pplns
			if (altA[i] != ppln1.allMinors[variantID[i]] && altA[i] != ppln1.allMajors[variantID[i]] 
					&& altA[i] != ppln2.allMinors[variantID[i]] && altA[i] != ppln2.allMajors[variantID[i]]) {
				//error("The alternative alleles for %s found is neither the minor nor the major allele in hapmap3.\n",variantID[i].c_str());
              printf("The alternative alleles for %ld found is neither the minor nor the major allele in hapmap3.\n",variantID[i]);
			} else{
				if ( altA[i] == ppln1.allMinors[variantID[i]] ) { // determin the f(alt in ppl1) 				
					f1 = ppln1MAF;
				} else if ( altA[i] == ppln1.allMajors[variantID[i]] ) {
					f1 = 1.0 - ppln1MAF;					
				} else {
					std::cout << "Alleles are not matched ppln1" << std::endl;
					/*
					std::cout << variantID[i].c_str() << std::endl;
					printf("altA[%d] = %s, ppln1.allMinor[%d] = %s, ppl2.allMajor[%d] = %s\n",i, altA[i].c_str(),i, ppln1.allMinors[i].c_str(),i, ppln2.allMajors[i].c_str());
					 */
				}
				if (altA[i] == ppln2.allMinors[variantID[i]]) {
					f2 = ppln2MAF;
				} else if (altA[i] == ppln2.allMajors[variantID[i]]){
					f2 = 1.0 - ppln2MAF;
				}	else {
					std::cout << "Alleles are not matched ppln2" << std::endl;
				}
		
			} 
			f12s.clear();
			f12s.push_back(f1);
			f12s.push_back(f2);
			// std::cout << "f1 = " << f1 << " f2 = " << f2 << std::endl;
			overlapAFs.push_back(f12s);
			
			// subset the overlapped gls
			overlapGLs.push_back(gllk[i]);
			overlapGDist.push_back(ppln1.geneticD[variantID[i]]);
			
			nSnpOverlapped ++ ;
		}
	}
	
	nOverlap = nSnpOverlapped;
	std::cout << "size of overlapGLs: " << overlapGLs.size() << std::endl;
	std::cout << "size of overlapAFs: " << overlapAFs.size() << std::endl;
	std::cout << "size of overlapVariantID: " << overlapVariantID.size() << std::endl;
	std::cout << "size of overlapGDist: " << overlapGDist.size() << std::endl;	
}


void GLFfromVCF::UpdateEandT(String &outFile){ // update the emission and transition probabilities
	// The order of the hidden states should be (0,0), (0,1), (1,0), (1,1)
	std::vector< std::pair <int,int> > states;
	std::pair<int,int> tmpState;
	for (int i = 0; i < numStatesOneChain; i ++) {
		for (int j = 0; j < numStatesOneChain; j ++) {			
			tmpState.first = i;
			tmpState.second = j;
			states.push_back(tmpState);
		}
	}
	/*printf("___state (size %d)___\n",(int) states.size());
	for (int i = 0; i < states.size(); i ++) {
		printf("state[%d] = (%d,%d)\n",i,states[i].first,states[i].second);
	}*/
	
	// Update the emission and transition probability
	e.Dimension(nOverlap, numStates, 0.0); 
	
	// genoF of (homozygous reference, heterozygous, non-reference homozygous)	
	// printf("overlapVariantID[0] = %s, overlapGLs[0][0] = %1.3e, overlapGLs[n][1] = %1.3e, overlapGLs[n][2] = %1.3e, overlapGLs[n][3] = %1.3e\n",
	//			 overlapVariantID[0].c_str(),overlapGLs[0][0], overlapGLs[0][1], overlapGLs[0][2], overlapGLs[0][3]);
	
	// printf("overlapAFs[0][0] = %1.3e, overlapAFs[0][1] = %1.3e\n",overlapAFs[0][0], overlapAFs[0][1]);
	/*
	for (int i = 0; i < nOverlap; i ++) {
		std::cout << overlapVariantID[i].c_str() << "\t" << overlapAFs[i][0] << "\t" << overlapAFs[i][1] << std::endl;
	} */
	
	for (int n = 0; n < nOverlap; n ++) {
		for (int i = 0; i < states.size(); i ++) {
			e[n][i] =  overlapGLs[n][0] * (1-overlapAFs[n][states[i].first]) * (1-overlapAFs[n][states[i].second]) // homo ref 0|0
								+ overlapGLs[n][1] * (1-overlapAFs[n][states[i].first]) * overlapAFs[n][states[i].second] // heter 0|1
								+ overlapGLs[n][1] * overlapAFs[n][states[i].first] * (1-overlapAFs[n][states[i].second]) // heter 1|0
								+ overlapGLs[n][2] * overlapAFs[n][states[i].first] * overlapAFs[n][states[i].second] // non-ref homo 1|1
							;
			// printf("e[%d][%d] = %f\n",n,i,e[n][i]);
		}
	}
	
	// Udpate the transition probability
	double pi = 0.0,pE = 0.2,pA; // pA,pE are prior probability of African and European ancestry
	pA = 1.0 - pE;
	std::vector< double > pEuropean; // prior for one chain: 0 - stands for European ancestry and 1 is for African ancestry	
	std::vector< double > statesPrior; // prior for double chain
	Matrix t0Tmp(numStatesOneChain,numStatesOneChain); // transition matrix for one chain
	Matrix tTmp(numStates,numStates); // full transition matrix at each locus (for the double chain)
	Matrix alpha, beta;
	alpha.Dimension(nOverlap, numStates,0.0);
	beta.Dimension(nOverlap, numStates,0.0);
	
	// prior probability for each state
	pEuropean.push_back(pE); pEuropean.push_back(pA);
	for (int i = 0; i < states.size(); i ++) {
		statesPrior.push_back(pEuropean[states[i].first]*pEuropean[states[i].second]);
		// printf("statePrior = %f\n",(double) statesPrior[i]);
	}

	// double dphyDist,dphyDist0;
	std::vector< double > dphyDist(nOverlap,0);
	StringArray distTokens; 
	// distTokens.AddTokens(overlapVariantID[0], ":");
	// std::cout << distTokens[1].c_str() << std::endl;
	// HMM forward-backward probability
	double paramShrinkage = 1;//1e-6;
	for (int n = 0; n < nOverlap; n ++) {
		// First, generate the transition probability for one chain
		//pi = 1 - exp(-overlapGDist[n]*(double)T);
		//std::cout << overlapVariantID[i].c_str() << std::endl;
		// distTokens.Clear(); // Get the genetic distance
		// distTokens.AddTokens(overlapVariantID[n], ":");
		// dphyDist[n] = distTokens[1].AsDouble();
		dphyDist[n] = (double) overlapVariantID[n];
		if (n > 0) {
			//pi = 1- exp(-4.0*11000*(overlapGDist[n+1]-overlapGDist[n])/(dphyDist[n]-dphyDist[0])*T); //11000 is the effective ppln size *1.0e6
			//pi = 1- exp(-4.0*11000*(overlapGDist[n+1]-overlapGDist[n])/100);
			// pi = exp(-4.0*11000*(overlapGDist[n]-overlapGDist[n-1])/(dphyDist[n]-dphyDist[n-1])/100*T);
            
            // using the recombination map
            pi = 1 - exp(-(overlapGDist[n+1]-overlapGDist[n])*0.01*T);

            // looks like that the T needs to be in the right scale 
            // using the relationship of physical distance and genetic distance
			// pi = (1- exp(-(dphyDist[n]-dphyDist[n-1])*1e-8*T))*paramShrinkage;

            // these are based on a different defnition of pis
			//pi = exp(-(overlapGDist[n+1]-overlapGDist[n])/(dphyDist[n+1]-dphyDist[n])*0.01*T); // this is the best one, but weird
             //pi = 0.5*(1-exp(-(overlapGDist[n+1]-overlapGDist[n])*0.01*T*2));
              
			// std::cout << "pi = " << pi << std::endl;
		}

		/*
		t0Tmp[0][0] = (1.0 + pi)/2.0;
		t0Tmp[0][1] = (1.0 - pi)/2.0;
		t0Tmp[1][0] = (1.0 - pi)/2.0;
		t0Tmp[1][1] = (1.0 + pi)/2.0;
		*/
		
		// std::cout << "pi = " << pi;
		// printf("A: %1.5e %1.5e %1.5e %1.5e\n", t0Tmp[0][0], t0Tmp[0][1], t0Tmp[1][0], t0Tmp[1][1]);
		
		// std::cout << pi << std::endl;
		/*t0Tmp[0][0] = 1.0 - pA+pi*pA; // = 1.0-0.8+pi*0.8 = 0.2+pi*0.8
		t0Tmp[0][1] = pA - pi*pA;
		t0Tmp[1][0] = pE - pi*pE;
		t0Tmp[1][1] = 1.0 - pE+pi*pE;*/

        t0Tmp[0][1] = pi*pA;
        t0Tmp[0][0] = 1 - t0Tmp[0][1];
        t0Tmp[1][0] = pi*pE;
        t0Tmp[1][1] = 1 - t0Tmp[1][0];
		 
		// printf("%1.5e %1.5e %1.5e %1.5e\n", t0Tmp[0][0], t0Tmp[0][1], t0Tmp[1][0], t0Tmp[1][1]);
		
		// Then, generate the final transition probabilty
		for	(int i = 0; i < states.size(); i ++){
			for	(int j = 0; j < states.size(); j ++){
				tTmp[i][j] = t0Tmp[states[i].first][states[j].first]*t0Tmp[states[i].second][states[j].second];
				// printf("tTmp[%d][%d] = %f\t",i,j,tTmp[i][j]);
			}
			// printf("\n");
		}
		t.push_back(tTmp);
	} // end loop over all nOverlap
	
	std::cout << "The size of emission probabilities: " << e.rows << " " << e.cols << std::endl;
	std::cout << "The size of transition probabilities: " << t.size() << std::endl;
	
	// Then forward and backward algorithm
	// calcualte alpha
	double c = 0.0;
	for (int n = 0; n < nOverlap; n ++) {
		c = 0.0;
		if (n == 0) { // initialize alpha_0(i) = p_i*b_i(o_1) where p_i is the prior probability of each states
			for (int i = 0; i < states.size(); i ++) {
				alpha[n][i] = statesPrior[i]*e[n][i];
				c += alpha[n][i];
				/*
				std::cout << " c = " << c << std::endl;
				std::cout << " alpha [ " << n << "][" << i << "] = " << alpha[n][i] << std::endl;
				 */
			}
			// rescale and update alpha into alpha/c 
			for (int i = 0; i < states.size(); i ++) {
				//printf("alpha[%d][%d] = %f\n",n,i,alpha[n][i]);
				alpha[n][i] /= c;
				//printf("alpha[%d][%d] = %f\n",n,i,alpha[n][i]);
			}
		} else {
			for (int i = 0; i < states.size(); i ++) {
				for (int j = 0; j < states.size(); j ++) {
					alpha[n][i] += alpha[n-1][j]*t[n-1][j][i]*e[n][i];
				}
				c += alpha[n][i];
			}
			/*
			if (n == nOverlap - 1) {
				std::cout << "alphas: " << alpha[n-1][0] << "," << t[n][0][0] << std::endl;
			}
			 */  // Why t[n] is not reasonable HERE?
			// printf("c[%d] = %1.3e\n",n,c);
			// rescale and update alpha into alpha/c
			for (int i = 0; i < states.size(); i ++) {
				alpha[n][i] /= c;
			}
		}			
	}
	
	// calcualte beta
	// initialize beta[nOverlap-1] = 1
	for (int n = nOverlap-1; n >= 0; n --) {
		if (n == nOverlap - 1) {
			for (int i = 0; i < states.size(); i ++) {
				beta[nOverlap-1][i] = 1.0;
			} 			
		} else {
			c = 0.0;
			for (int i = 0; i < states.size(); i ++) {
				for (int j = 0; j < states.size(); j ++) {
					beta[n][i] += beta[n+1][j]*t[n][i][j]*e[n+1][j];
					//printf( "beta[%d][%d] = %1.3e, alpha = %1.3e, t[%d][%d][%d] = %1.3e, e[%d][%d] = %1.3e\n", 
					//			 n+1, i, beta[n+1][i], alpha[n][i], n, i, j, t[n][i][j], n+1, j, e[n+1][j] );
				}
				c += beta[n][i];
			}
			for (int i = 0; i < states.size(); i ++) {
				beta[n][i] /= c;
			}
		} 		
	}
	

	/*for (int n = 0; n < nOverlap; n ++) {
		for (int i = 0; i < states.size(); i ++) {
			printf("beta[%d][%d] = %1.3e, alpha[%d][%d] = %1.3e\t",n,i,beta[n][i],n,i,alpha[n][i]);
		}
		printf("\n");
	}*/

	// calcualte Pr(q_t = i|o,lambda) = alpha_t(i)beta_t(i)/sum(alpha_t(i)beta_t(i))
	std::vector<double> totalP(nOverlap,0);
	std::vector< std::vector<double> > condP;
	for (int n = 0; n < nOverlap; n ++) {
		for (int i = 0; i < states.size(); i ++) {
			totalP[n] += alpha[n][i]*beta[n][i];
		}
	}
	
	std::vector<double> tmpCondP;
	for (int n = 0; n < nOverlap; n ++) {
		tmpCondP.clear();
		for (int i = 0; i < states.size(); i ++) {
			tmpCondP.push_back(alpha[n][i]*beta[n][i]/totalP[n]);
			/*if (n == 0) {
				//std::cout << alpha[n][i] << "\t" << beta[n][i] << "\t" << totalP[n] << std::endl;
				printf("totalP[%d] = %1.3e\n",n,totalP[n]);
			}*/
		}
		condP.push_back(tmpCondP);
	}
	
	std::cout << std::endl;
	std::cout << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
	std::cout << "size of alpha = " << alpha.rows << " " << alpha.cols << std::endl;
	std::cout << "size of beta = " << beta.rows << " " << beta.cols << std::endl;
	std::cout << "size of condP = " << condP.size() << std::endl;
	
	// output the conditional probabilities
	FILE *outF = fopen( (outFile+idSubj2Decomp+".txt").c_str(),"w");
    if (outF == NULL){
        error("Cannot open file (%s) to write\n",(outFile+idSubj2Decomp+".txt").c_str());
    }   

	std::cout << "Writing result to file " << (outFile+idSubj2Decomp+".txt").c_str() << std::endl;

    StringArray chrPos;
    for (int n = 0; n < nOverlap; n ++) {
      //std::cout << overlapVariantID[n].c_str() << "\t";
      //chrPos.ReplaceTokens(overlapVariantID[n], ":");
      //fprintf(outF, "%s\t%s\t",chrPos[0].c_str(),chrPos[1].c_str());
			fprintf(outF, "%ld\t", overlapVariantID[n]);
			
      for (int i = 0; i < states.size(); i ++) {
        //std::cout << condP[n][i] << "\t";
        fprintf(outF, "%1.3e\t",condP[n][i]);
      }
      //std::cout << std::endl;
      fprintf(outF, "\n");
    }
    std::cout << "Done!" << std::endl;  

    return;
}
