/*
 *  LocalAncestry.cpp
 *  LocalAncestry
 *
 *  Created by Youna Hu on 3/31/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "LocalAncestry.h"
#include "Parameters.h"
#include "Error.h"
#include "StringArray.h"
#include "RefPplns.h"
#include "GLFfromVCF.h"

int main(int argc, char **argv){
	ParameterList pl;
	// Should be a vcf with multiple subjects, but we will look at one subject for now
	String testFile;
	// Hapmap vcf is read in to subsetting the allele frequency
	// Note: need to see the difference in using the AF and empirical MAF (.frq file)
	String freqCEU;
	String freqYRI;
	String geneticDistance;
	String outAncestryF;
	String gllkKey = "GL";
	int colSubject = 9;
	int Time = 6;

    printf("SEQMIX 0.1 -- local ancestry inference for SEQuenced adMIXed individuals\n");
    printf("(c) 2013 Youna Hu, Goncalo Abecasis\n");

	BEGIN_LONG_PARAMETERS(options)
		LONG_PARAMETER_GROUP("I/O files")
		LONG_STRINGPARAMETER("fileToDecompose", &testFile)
		LONG_STRINGPARAMETER("freqCEUfile", &freqCEU)	
		LONG_STRINGPARAMETER("freqYRIfile", &freqYRI)	
		LONG_STRINGPARAMETER("geneticDistfile", &geneticDistance)
		LONG_STRINGPARAMETER("outputAncestryF", &outAncestryF)
	  LONG_PARAMETER_GROUP("Input Parameters")
		LONG_STRINGPARAMETER("GLLkey", &gllkKey)
		LONG_INTPARAMETER("T(Admixed Time)", &Time)
		LONG_INTPARAMETER("SubjCol", &colSubject)
	END_LONG_PARAMETERS();
	
	pl.Add( (new LongParameters("Local ancestry decomposition:", options) ) -> SetPrecision(3));
	pl.Read(argc, argv);
	pl.Status();
	/*	
	RefPplns testEA;
	testEA.UpdateMAFs(freqCEU);
	RefPplns testAF;
	testAF.UpdateMAFs(freqYRI);
	 */
    
    // do this in version 0.2
    // colSubject += 8; /*since the individual information starts at the 9th column, we hard cold it here */

	GLFfromVCF testGLF(testFile,gllkKey,colSubject,Time);
	
	testGLF.UpdateGLFs(testFile);
	testGLF.UpdateGenoFreqs(freqCEU,freqYRI,geneticDistance);
	testGLF.UpdateEandT(outAncestryF);
	return 0;
}
