/*
 *  RefPplns.cpp
 *  LocalAncestry
 *
 *  Created by Youna Hu on 3/31/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "RefPplns.h"

RefPplns::RefPplns(){
	colChr = 0;
	colPos = 1;
	colMinor = 3;
	colMajor = 4;
	colMAF = 5;
}

void RefPplns::UpdateMAFs(String inFile){
	FILE *HapMapF = fopen(inFile.c_str(), "r");
	if (HapMapF == NULL){
		error("Error in opening %s!\n",inFile.c_str());
		return ;
	}
	std::cout << "reading file " << inFile.c_str() << std::endl;
	int fileStep = 0;
	String fileEntry;
	StringArray fileTokens;
	long snpID;
	// read in the file and update the SNP frequencies and minor alleles
	while (!feof(HapMapF)) {
		fileEntry.ReadLine(HapMapF);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",inFile.c_str(),fileStep);
		}
		fileTokens.AddTokens(fileEntry, "\t");
		if (fileTokens[colChr] != "CHR") { // if the line is not the header
		//if (fileStep != 0){ // if the line is the header (we assume that there is header to the file
			//snpID = fileTokens[colChr] + ":" + fileTokens[colPos];
			snpID = fileTokens[colPos].AsInteger();
			allSNPfrq[snpID] = fileTokens[colMAF].AsDouble();
			if ( allSNPfrq[snpID] < 0.005 ) allSNPfrq[snpID] = 0.005;
			if ( allSNPfrq[snpID] > 0.995 ) allSNPfrq[snpID] = 0.995;
			allMinors[snpID] = fileTokens[colMinor][0];
			allMajors[snpID] = fileTokens[colMajor][0];
			// problem here, how to use map properly?
			//allSNPfrq.insert(std::map<String,double>::value_type(snpID,fileTokens[colMAF].AsDouble()));
			//allMinors.insert(std::map<String,double>::value_type(snpID,fileTokens[colMinor]));
			// std::cout << snpID.c_str() << " Allele frequency: " << fileTokens[colMAF].AsDouble() << std::endl;
			/*
			if (snpID == "20:61795") {
				std::cout << "Ppl " << this << std::endl;
				std::cout << "AllMinors = " << allMinors[snpID].c_str() << "," << "AllMajors = " << allMajors[snpID].c_str() << std::endl;
			}
			 */
		}
		fileEntry.Clear();
		fileTokens.Clear();
		fileStep ++;
	}
	fclose(HapMapF);
	
	if (allSNPfrq.size() != allMinors.size()) {
		error("Error: The size of allele frequency vector should be the same as that of minor allele vector!\n");
	} else {
		size = allSNPfrq.size();
	}
	printf("The number of refefence SNPs is %d.\n",(int) allSNPfrq.size());
}

void RefPplns::UpdateGDist(String inFile){ // here, for simplicity, I just use the genetic distance for chromosome 20 obtained from IMPUTE, this will be estimated by the data in the future
	FILE *HapMapF = fopen(inFile.c_str(), "r");
	if (HapMapF == NULL){
		error("Error in opening %s!\n",inFile.c_str());
		return ;
	} 
	std::cout << "reading file " << inFile.c_str() << std::endl;

	int fileStep = 0;
	String fileEntry;
	StringArray fileTokens;
	long snpID;
	// read in the file and update the SNP frequencies and minor alleles
	while (!feof(HapMapF)) {
		fileEntry.ReadLine(HapMapF);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",inFile.c_str(),fileStep);
		}
		fileTokens.AddTokens(fileEntry, "\t");		
		snpID = fileTokens[0].AsInteger();
		geneticD[snpID] = fileTokens[1].AsDouble();
		fileEntry.Clear();
		fileTokens.Clear();
	}
	fclose(HapMapF);
}
